package ch.inf3.horizoncut.cli;

import ch.inf3.horizoncut.data.DisplayCalculator;
import ch.inf3.horizoncut.data.DistanceLayer;
import ch.inf3.horizoncut.data.Feature;
import ch.inf3.horizoncut.data.ObservedFeature;
import ch.inf3.horizoncut.data.Position;
import ch.inf3.horizoncut.data.Tile;
import ch.inf3.horizoncut.data.TileMap;
import ch.inf3.horizoncut.data.TileReader;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.NavigableMap;
import java.util.TreeMap;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.imageio.ImageIO;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class Cli {

    private static final Logger LOG = LoggerFactory.getLogger(Cli.class);

    private static final int SLICE_DISTANCE = 100;

    public static void main(String[] args) throws Exception {

        long initStart = System.currentTimeMillis();
        Position eyePosition = new Position(46.326570, 7.466880);
        TileMap tileMap = new TileReader("/home/valdor/Downloads/SRTM1/").readHGTFiles(eyePosition, 2);

        final int width = 8000;
        final int height = 2000;

        Tile atOrigin = tileMap.getByLatLon(eyePosition.gridLat, eyePosition.gridLon);
        double eyeHeight = atOrigin.elevation(eyePosition.row, eyePosition.col);

        double minHorizAngle = 107; //107; // 135 ~Brunegghorn
        double maxHorizAngle = 226; //226; // 197 ~Combin de Valsorey

        DisplayCalculator dc = new DisplayCalculator(width, height, minHorizAngle, maxHorizAngle, eyePosition, eyeHeight);

        int maxDistance = 100_000;
        int minDistance = 500;

        Collection<Feature> rawFeatures = new ArrayList<>();
        rawFeatures.addAll(Feature.readFeatures("data/osm-overpass-peaks.geojson", tileMap, 150, true));
        rawFeatures.addAll(Feature.readFeatures("data/osm-overpass-villages.geojson", tileMap, height - 50, false));

        Collection<ObservedFeature> allFeatures = rawFeatures.stream()
                .map(f -> f.observe(dc))
                .sorted()
                .collect(Collectors.toList());

        BufferedImage bufferedImg = new BufferedImage(width, height + 1, BufferedImage.TYPE_INT_RGB);
        Graphics2D g = (Graphics2D) bufferedImg.getGraphics();

        long initStop = System.currentTimeMillis();
        LOG.info("Initialized system for computations from {}@{}m with distances of {} to {} ({} ms)",
                eyePosition.getLatLon(), eyeHeight, minDistance, maxDistance, initStop - initStart);

        long layersStart = System.currentTimeMillis();
        final NavigableMap<Integer, DistanceLayer> dLayers = new TreeMap<>();
        
        for (int d = maxDistance; d > minDistance; d -= SLICE_DISTANCE) {
            final int cDist = d;
            double[] data = computeAtDistance(tileMap, cDist, dc);
            dLayers.put(d, new DistanceLayer(d - SLICE_DISTANCE, d, data));
        }

        long layersEnd = System.currentTimeMillis();
        LOG.info("Computed {} layers ({} ms)", dLayers.size(), layersEnd - layersStart);

        long horizonStart = System.currentTimeMillis();
        // We compute the horizon and delete any layer that is behind the horizon
        // and is not used at all for the landscape.
        // For this, we are using the fact that we computed the layers starting
        // with the farthest ones in the build loop.
        DistanceLayer horizonLayer = dLayers.values().stream()
                .reduce((a, b) -> a.mergeWith(b))
                .get();
        
        LOG.info("Horizon is from {} to {}", horizonLayer.nearDistance, horizonLayer.farDistance);
        
        Integer fKey = dLayers.firstKey();
        Integer lKey = dLayers.ceilingKey(horizonLayer.farDistance);
        
        LOG.info("Split horizon from {} ({}) to {} ({})", fKey, fKey, lKey, horizonLayer.farDistance);
        final NavigableMap<Integer, DistanceLayer> hLayers = dLayers.subMap(fKey, true, lKey, true);

        // We recompute the min and max distances after the cleanup of layers that
        // are not useful for the display.
        maxDistance = hLayers.values().stream()
                .map(l -> l.farDistance)
                .max(Integer::compareTo)
                .get();

        minDistance = hLayers.values().stream()
                .map(l -> l.nearDistance)
                .min(Integer::compareTo)
                .get();

        dc.updateMinMaxVertAngle(hLayers.values());

        long horizonEnd = System.currentTimeMillis();
        LOG.info("There are {} layers remaining after horizon cleanup at distances {} to {} ({} ms)",
                hLayers.size(), minDistance, maxDistance, horizonEnd - horizonStart);

        long labelsStart = System.currentTimeMillis();
        
        // We start processing the features to avoid overlapping features in the
        // labels as well as displaying features that are hidden from the observation
        // point.

        Predicate<ObservedFeature> isVisibleFilter = (feature) -> {
            int x = dc.getXAtBearing(feature.bearing);
            return DistanceLayer.isVisible(hLayers.values(), feature.distance, x, feature.visibleAngle);
        };

        TreeMap<Integer, ObservedFeature> labeledFeatures = allFeatures.stream()
                .filter(f -> dc.isDisplayed(f.bearing, f.visibleAngle))
                .filter(isVisibleFilter)
                .collect(Collectors.toMap(f -> f.getObservedX(), f -> f, (a, b) -> Feature.highest(a, b), TreeMap::new));

        Integer cCol = labeledFeatures.firstKey();
        while (cCol < labeledFeatures.lastKey()) {
            Integer nCol = labeledFeatures.higherKey(cCol);

            if (nCol - cCol <= 25) {
                double nScore = labeledFeatures.get(nCol).getScore();
                double cScore = labeledFeatures.get(cCol).getScore();

                if (nScore > cScore) {
                    LOG.debug("Removing {} ({}) to keep {} ({}) at X {}", labeledFeatures.get(cCol), cScore, labeledFeatures.get(nCol), nScore, cCol);
                    labeledFeatures.remove(cCol);
                    cCol = nCol;
                } else {
                    labeledFeatures.remove(nCol);
                }
            } else {
                cCol = nCol;
            }
        }

        for (ObservedFeature feature : labeledFeatures.values()) {
            var featureLayer = hLayers.floorEntry(feature.distance);
            if (featureLayer == null) {
                LOG.info("Unable to find layer for feature {}", feature);
                continue;
            }

            featureLayer.getValue().addFeature(feature);
        }

        long labelsEnd = System.currentTimeMillis();
        LOG.info("Computed {} visible labels ({} ms)", labeledFeatures.size(), labelsEnd - labelsStart);

        long mergeStart = System.currentTimeMillis();
        List<DistanceLayer> sortedLayers = new ArrayList<>();
        Integer cDist = hLayers.firstKey();

        DistanceLayer cLayer = hLayers.get(cDist);
        while (cDist < hLayers.lastKey()) {
            Integer nDist = hLayers.higherKey(cDist);
            DistanceLayer nLayer = hLayers.get(nDist);

            var bothFeatures = Stream.concat(cLayer.getHorizonFeatures().stream(),nLayer.getHorizonFeatures().stream())
                    .filter(isVisibleFilter)
                    .collect(Collectors.toList());

            DistanceLayer mLayer = cLayer.mergeWith(nLayer);
            var mergeFeatures = nLayer.getHorizonFeatures().stream()
                    .filter(isVisibleFilter)
                    .collect(Collectors.toList());

            boolean continueMerge = false;

            if (mergeFeatures.isEmpty()) {
                // If no features are present on the merge result, we always continue;
                continueMerge = true;
            } else {
                double bothScore = bothFeatures.stream()
                        .map(Feature::getScore)
                        .reduce(Double::sum)
                        .orElse(0.0);
                double mergeScore = mergeFeatures.stream()
                        .map(Feature::getScore)
                        .reduce(Double::sum)
                        .orElse(0.0);

                var missingFeatures = bothFeatures.stream()
                        .filter(f -> !mergeFeatures.contains(f))
                        .collect(Collectors.toList());

                LOG.info("Merging layers {} and {} ({}) would generate {} feature score: {}",
                        cLayer.nearDistance, nLayer.nearDistance, bothScore,
                        mergeScore, continueMerge);

                if (!missingFeatures.isEmpty()) {
                    LOG.info("  Removed {}", missingFeatures);
                }

                continueMerge = mergeScore >= bothScore;
            }

            if (!continueMerge) {
                sortedLayers.add(cLayer);
                cLayer = nLayer;
            } else {
                cLayer = mLayer;
            }

            cDist = nDist;
        }

        sortedLayers.add(cLayer);

        long mergeEnd = System.currentTimeMillis();
        LOG.info("There are {} layers remaining after layer merges for {} features ({} ms)",
                sortedLayers.size(), labeledFeatures.size(), mergeEnd - mergeStart);

        long renderStart = System.currentTimeMillis();
        // We loop in descending order to avoid writing further points on top
        // of nearer points.
        for (DistanceLayer layer : sortedLayers) {
            float dPos = (float) ((maxDistance - layer.nearDistance) / (maxDistance * 1.0f));
            Color layerColor = Color.getHSBColor(dPos * 1f, 1.0f, 1.0f);

            g.setColor(layerColor);

            Polygon p = new Polygon();
            p.addPoint(0, height);
            for (int x = 0; x < width; x++) {
                double visibleAngle = layer.data[x];
                int y = dc.getYAtAngle(visibleAngle);

                p.addPoint(x, y);
            }

            p.addPoint(width, height);
            g.drawPolygon(p);

            for (ObservedFeature feature : layer.getFeatures()) {
                if (!dc.isDisplayed(feature.bearing, feature.visibleAngle)) {
                    LOG.debug("Feature {} is not visible at bearing {} and angle {}", feature.name, feature.bearing, feature.visibleAngle);
                    continue;
                }

                int x = dc.getXAtBearing(feature.bearing);
                int y = dc.getYAtAngle(feature.visibleAngle);

                boolean isVisible = DistanceLayer.isVisible(hLayers.values(), feature.distance, x, feature.visibleAngle);
                if (!isVisible) {
                    LOG.debug("Feature {} is hidden behind terrain", feature.name);
                    continue;
                }

                Shape s = new Rectangle(x -5 , y - 5, 10, 10);

                if (layer.isAtHorizon(feature)) {
                    g.fill(s);
                } else {
                    g.draw(s);
                }
            }
        }

        Font font = new Font("Helvetica", Font.BOLD, 12);
        g.setColor(Color.white);
        g.setFont(font);

        for (ObservedFeature feature : labeledFeatures.values()) {
            if (!dc.isDisplayed(feature.bearing, feature.visibleAngle)) {
                LOG.debug("Feature {} is not visible at bearing {} and angle {}", feature.name, feature.bearing, feature.visibleAngle);
                continue;
            }

            int x = dc.getXAtBearing(feature.bearing);

            boolean isVisible = DistanceLayer.isVisible(hLayers.values(), feature.distance, x, feature.visibleAngle);
            if (!isVisible) {
                LOG.debug("Feature {} is not visible, hidden behind terrain.", feature.name);
                continue;
            }

            int y = dc.getYAtAngle(feature.visibleAngle);

            LOG.debug("Peak {} is at bearing {} ({}x{})", feature.name, feature.bearing, x, y);

            String title = String.format("%s (%.0f)", feature.name, feature.getScore());

            AffineTransform old = g.getTransform();
            g.rotate(Math.toRadians(-45), x, feature.textPosition);
            g.drawString(title, x, feature.textPosition);
            g.setTransform(old);

            g.drawLine(x, y, x, feature.textPosition);
        }

        g.setColor(Color.gray);
        for (int i = 0; i < 360; i += 1) {
            int x = dc.getXAtBearing(i);

            g.drawLine(x, 0, x, height);
            g.drawString(i + "", x + 5, 20);
        }

        long renderEnd = System.currentTimeMillis();
        LOG.info("Rendered panorama ({} ms)", renderEnd - renderStart);

        g.dispose();

        try {
            String imageName = String.format("image-%05d.png", 1);
            File outputfile = new File("/home/valdor/", imageName);
            outputfile.createNewFile();

            ImageIO.write(bufferedImg, "png", outputfile);
        } catch (IOException ex) {
            LOG.error("Exception while writing image", ex);
        }
    }

    static final double[] computeAtDistance(TileMap tileMap, int d, DisplayCalculator dc) {

        double[] output = new double[dc.getWidth()];

        for (int x = 0; x < dc.getWidth(); x += 1) {
            double bearing = dc.getBearingAtX(x);
            Position target = Position.get(dc.getEyePosition(), Math.toRadians(bearing), d);

            Tile atPos = tileMap.getByLatLon(target.gridLat, target.gridLon);
            double elevation = atPos.elevation(target.row, target.col);

            int dMeters = dc.getEyePosition().distanceToMeters(target);
            double visibleAngle = dc.calculateVisibilityAngle(elevation, dMeters);

            output[x] = visibleAngle;
        }

        return output;
    }

}
