package ch.inf3.horizoncut.data;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class DistanceLayer implements Comparable<DistanceLayer>, Cloneable {

    private static final Logger LOG = LoggerFactory.getLogger(DistanceLayer.class);

    public final int nearDistance;
    public final int farDistance;
    public final double[] data;

    public final Collection<ObservedFeature> horizonFeatures;
    public final Collection<ObservedFeature> nonHorizonFeatures;

    public DistanceLayer(int nearDistance, int farDistance, double[] data) {
        this(nearDistance, farDistance, data, new ArrayList<>(), new ArrayList<>());
    }

    private DistanceLayer(int nearDistance, int farDistance, double[] data,
            Collection<ObservedFeature> horizonFeatures, Collection<ObservedFeature> nonHorizonFeatures) {
        this.nearDistance = nearDistance;
        this.farDistance = farDistance;
        this.data = data;

        this.horizonFeatures = horizonFeatures;
        this.nonHorizonFeatures = nonHorizonFeatures;
    }

    @Override
    public int compareTo(DistanceLayer other) {
        return Integer.compare(this.nearDistance, other.nearDistance);
    }

    @Override
    protected DistanceLayer clone() {
        double[] cloneData = Arrays.copyOf(data, data.length);

        Collection<ObservedFeature> horizonFeatures = new ArrayList<>(this.horizonFeatures);
        Collection<ObservedFeature> nonHorizonFeatures = new ArrayList<>(this.nonHorizonFeatures);

        return new DistanceLayer(nearDistance, farDistance, cloneData, horizonFeatures, nonHorizonFeatures);
    }

    public DistanceLayer mergeWith(DistanceLayer other) {
        DistanceLayer near = this.compareTo(other) <= 0 ? this : other;
        DistanceLayer far = this.compareTo(other) > 0 ? this : other;

        double[] output = new double[near.data.length];

        int countNear = 0;
        int countFar = 0;

        for (int d = 0; d < near.data.length; d += 1) {
            double yNear = near.data[d];
            double yFar = far.data[d];

            if (yNear >= yFar) {
                output[d] = yNear;
                countNear += 1;
            } else {
                output[d] = yFar;
                countFar += 1;
            }
        }
        
        int tolerance = near.data.length / 100;

        int outputNearDistance = (countNear > tolerance) ? near.nearDistance : far.nearDistance;
        int outputFarDistance = (countFar > tolerance) ? far.farDistance : near.farDistance;
        
        LOG.debug("Merging {}-{} ({}) with {}-{} ({}) {}",
                near.nearDistance, near.farDistance, countNear,
                far.nearDistance, far.farDistance, countFar,
                tolerance);

        List<ObservedFeature> outHorizonFeatures = Stream.concat(this.horizonFeatures.parallelStream(), other.horizonFeatures.parallelStream())
                .filter(f -> isAtHorizon(f, output))
                .collect(Collectors.toList());

        List<ObservedFeature> outNonHorizonFeatures = new ArrayList<>();
        outNonHorizonFeatures.addAll(this.nonHorizonFeatures);
        outNonHorizonFeatures.addAll(other.nonHorizonFeatures);

        return new DistanceLayer(outputNearDistance, outputFarDistance, output, outHorizonFeatures, outNonHorizonFeatures);
    }

    /**
     * Returns true if the given layer is always lower than the current layer.
     *
     * @param other
     * @return
     */
    public boolean isLower(DistanceLayer other) {
        for (int d = 0; d < this.data.length; d += 1) {
            double yThis = this.data[d];
            double yOther = other.data[d];

            if (yThis == 0 || yOther == 0) {
                continue;
            }

            if (yOther >= yThis) {
                return false;
            }
        }

        return true;
    }

    public void addFeature(ObservedFeature feature) {
        if (feature.requireHorizon) {
            this.horizonFeatures.add(feature);
        } else {
            this.nonHorizonFeatures.add(feature);
        }
    }

    public Collection<ObservedFeature> getHorizonFeatures() {
        return Collections.unmodifiableCollection(this.horizonFeatures);
    }

    public Collection<ObservedFeature> getFeatures() {
        return Stream.concat(this.horizonFeatures.stream(), this.nonHorizonFeatures.stream())
                .sorted()
                .collect(Collectors.toList());
    }

    public boolean isAtHorizon(ObservedFeature feature) {
        return isAtHorizon(feature, data);
    }

    private static boolean isAtHorizon(ObservedFeature feature, double[] data) {
        int xFeature = feature.getObservedX();
        double angleFeature = feature.visibleAngle;

        double angleThis = data[xFeature];

        boolean isAtHorizon = Math.abs(angleThis - angleFeature) < 0.2;

        LOG.debug("Feature {}, {} vs {} -> Horizon {}", feature.name, angleFeature, angleThis, isAtHorizon);

        return isAtHorizon;
    }

    public int countHorizonFeatures() {
        return this.horizonFeatures.size();
    }

    public double getMax() {
        return Arrays.stream(this.data)
                .boxed()
                .filter(e -> !e.isNaN())
                .max(Double::compareTo)
                .orElse(0.0);
    }

    public static boolean isVisible(Collection<DistanceLayer> sortedLayers, int distance, int x, double visibleAngle) {
        for (DistanceLayer l : sortedLayers) {
            if (l.farDistance > (distance - 200)) {
                return true;
            }

            double c = l.data[x];

            if (c >= visibleAngle) {
                return false;
            }
        }

        return true;
    }

}
