package ch.inf3.horizoncut.data;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;

public class Feature {

    private final Position position;
    private final double elevation;

    private final double score;

    public final String name;

    public final int textPosition;
    public final boolean requireHorizon;

    protected Feature(Position position, double elevation, String name, int textPosition, boolean requireHorizon, double score) {
        this.position = position;
        this.elevation = elevation;
        this.name = name;
        this.textPosition = textPosition;
        this.requireHorizon = requireHorizon;
        this.score = score;
    }

    public ObservedFeature observe(DisplayCalculator observation) {
        return new ObservedFeature(this.position, this.elevation, observation, this.name, this.textPosition, this.requireHorizon, this.score);
    }

    public static Collection<Feature> readFeatures(String fileName, TileMap tileMap, int textPosition, boolean requireHorizon) throws IOException {
        Collection<Feature> peaks = new ArrayList<>();

        try (FileReader fr = new FileReader(fileName)) {
            Gson gson = new GsonBuilder().create();
            GeoJsonExport export = gson.fromJson(fr, GeoJsonExport.class);

            for (GeoJsonExport.Feature feature : export.features) {
                double[] coordinates = feature.geometry.coordinates;
                String name = feature.properties.name;

                if (name == null || name.isBlank()) {
                    continue;
                }

                Position featurePosition = new Position(coordinates[1], coordinates[0]);

                Tile atPos = tileMap.getByLatLon(featurePosition.gridLat, featurePosition.gridLon);
                double elevation = atPos.elevation(featurePosition.row, featurePosition.col);

                double score = elevation;
                if (feature.properties.prominence != null) {
                    score += 10 * Integer.parseInt(feature.properties.prominence);
                }
                if (feature.properties.wikidata != null) {
                    score += 9_000;
                }

                Feature peak = new Feature(featurePosition, elevation, name, textPosition, requireHorizon, score);
                peaks.add(peak);
            }
        }

        return peaks;
    }

    public static <T extends Feature> T highest(T... features) {
        return Arrays.stream(features)
                .max((a, b) -> Double.compare(a.getScore(), b.getScore()))
                .get();
    }

    public double getScore() {
        return this.score;
    }

    @Override
    public String toString() {
        return name + "(score=" + score + ")";
    }

}
